import datetime
from requests import ConnectionError

from .baseendpoint import BaseEndpoint
from ..resources import LoginResource
from ..exceptions import (
    LoginError,
    APIError,
    InvalidResponse
)
from ..utils import check_status_code


class Login(BaseEndpoint):
    """
    Login operations.
    """

    _error = LoginError

    def __call__(self, session=None, lightweight=None):
        """
        Makes login request.

        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: LoginResource
        """
        (response, elapsed_time) = self.request(self.url, session=session)
        self.client.set_session_token(response.get('sessionToken'))
        return self.process_response(response, LoginResource, elapsed_time, lightweight)

    def request(self, method=None, params=None, session=None):
        session = session or self.client.session
        date_time_sent = datetime.datetime.utcnow()
        try:
            response = session.post(self.url, data=self.data, headers=self.client.login_headers, cert=self.client.cert)
        except ConnectionError:
            raise APIError(None, exception='ConnectionError')
        except Exception as e:
            raise APIError(None, exception=e)
        elapsed_time = (datetime.datetime.utcnow() - date_time_sent).total_seconds()

        check_status_code(response)
        try:
            response_data = response.json()
        except ValueError:
            raise InvalidResponse(response.text)

        if self._error_handler:
            self._error_handler(response_data)
        return response_data, elapsed_time

    def _error_handler(self, response, method=None, params=None):
        if response.get('loginStatus') != 'SUCCESS':
            raise self._error(response)

    @property
    def url(self):
        return '%s%s' % (self.client.identity_cert_uri, 'certlogin')

    @property
    def data(self):
        return {'username': self.client.username, 'password': self.client.password}



import unittest
from tests import mock

from betfairlightweight import APIClient
from betfairlightweight.endpoints import Streaming


class StreamingTest(unittest.TestCase):

    def setUp(self):
        self.client = APIClient('username', 'password', 'app_key', 'UK')
        self.streaming = Streaming(self.client)

    def test_init(self):
        self.assertEqual(self.streaming.client, self.client)

    @mock.patch('betfairlightweight.endpoints.streaming.BetfairStream')
    def test_create_stream(self, mock_betfair_stream):
        response = self.streaming.create_stream(1, 2, 6, 1024, 'TestSocket')

        assert mock_betfair_stream.call_count == 1
        mock_betfair_stream.assert_called_with(1, 2, app_key=self.streaming.client.app_key,
                                               session_token=self.streaming.client.session_token, timeout=6,
                                               buffer_size=1024, description='TestSocket', host=None)
        assert response == mock_betfair_stream()

    @mock.patch('betfairlightweight.endpoints.streaming.HistoricalStream')
    def test_create_historical_stream(self, mock_stream):
        dir = 'test'
        listener = mock.Mock()
        self.streaming.create_historical_stream(dir, listener)

        listener.register_stream.assert_called_with('HISTORICAL', 'marketSubscription')
        mock_stream.assert_called_with(dir, listener)

    @mock.patch('betfairlightweight.endpoints.streaming.HistoricalGeneratorStream')
    def test_create_historical_generator_stream(self, mock_stream):
        dir = 'test'
        listener = mock.Mock()
        self.streaming.create_historical_generator_stream(dir, listener)

        listener.register_stream.assert_called_with('HISTORICAL', 'marketSubscription')
        mock_stream.assert_called_with(dir, listener)



import logging

from .apiclient import APIClient
from .exceptions import BetfairError
from .streaming import StreamListener
from . import filters

__title__ = 'betfairlightweight'
__version__ = '1.10.3'
__author__ = 'Liam Pauling'

# Set default logging handler to avoid "No handler found" warnings.
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



print("import: 'betfairlightweight'")
import betfairlightweight

print("import: 'betfairlightweight.endpoints'")
import betfairlightweight.endpoints

print("import: 'betfairlightweight.resources'")
import betfairlightweight.resources

print("import: 'betfairlightweight.streaming'")
import betfairlightweight.streaming




import unittest
import socket
import time
import threading
from tests import mock

from betfairlightweight.streaming.betfairstream import (
    BetfairStream,
    HistoricalStream,
    HistoricalGeneratorStream,
)
from betfairlightweight.exceptions import (
    SocketError,
    ListenerError,
)


class BetfairStreamTest(unittest.TestCase):

    def setUp(self):
        self.mock_listener = mock.Mock()
        self.mock_listener.on_data.return_value = False
        self.unique_id = 1
        self.app_key = 'app_key'
        self.session_token = 'session_token'
        self.timeout = 6
        self.buffer_size = 1024
        self.description = 'test_stream'
        self.betfair_stream = BetfairStream(self.unique_id, self.mock_listener, self.app_key, self.session_token,
                                            self.timeout, self.buffer_size, self.description, None)

    def test_init(self):
        assert self.betfair_stream._unique_id == self.unique_id
        assert self.betfair_stream.listener == self.mock_listener
        assert self.betfair_stream.app_key == self.app_key
        assert self.betfair_stream.session_token == self.session_token
        assert self.betfair_stream.timeout == self.timeout
        assert self.betfair_stream.buffer_size == self.buffer_size
        assert self.betfair_stream.description == self.description
        assert self.betfair_stream.host == 'stream-api.betfair.com'
        assert self.betfair_stream.receive_count == 0
        assert self.betfair_stream.datetime_last_received is None

        assert self.betfair_stream._socket is None
        assert self.betfair_stream._running is False

    def test_host_init(self):
        betfair_stream = BetfairStream(self.unique_id, self.mock_listener, self.app_key, self.session_token,
                                       self.timeout, self.buffer_size, self.description, 'integration')
        assert betfair_stream.host == 'stream-api-integration.betfair.com'

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream.authenticate')
    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._connect')
    @mock.patch('betfairlightweight.streaming.betfairstream.threading')
    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._read_loop')
    def test_start(self, mock_read_loop, mock_threading, mock_connect, mock_authenticate):
        self.betfair_stream._running = True
        self.betfair_stream.start()
        mock_read_loop.assert_called_with()

        self.betfair_stream.start(async_=True)
        mock_threading.Thread.assert_called_with(name=self.description, target=mock_read_loop)

        self.betfair_stream._running = False
        self.betfair_stream.start(async_=False)
        mock_connect.assert_called_with()
        mock_authenticate.assert_called_with()

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._create_socket')
    def test_connect(self, mock_create_socket):
        self.betfair_stream._connect()

        assert self.betfair_stream._running is True
        mock_create_socket.assert_called_with()

    def test_stop(self):
        self.betfair_stream.stop()
        assert self.betfair_stream._running is False
        assert self.betfair_stream._socket is None

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._send')
    def test_authenticate(self, mock_send):
        self.betfair_stream.authenticate()
        mock_send.assert_called_with(
                {'id': self.betfair_stream._unique_id, 'appKey': self.app_key, 'session': self.session_token, 'op': 'authentication'}
        )

        self.betfair_stream.authenticate()
        mock_send.assert_called_with(
                {'id': self.betfair_stream._unique_id, 'appKey': self.app_key, 'session': self.session_token, 'op': 'authentication'}
        )

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._send')
    def test_heartbeat(self, mock_send):
        self.betfair_stream.heartbeat()
        mock_send.assert_called_with(
                {'id': self.betfair_stream._unique_id, 'op': 'heartbeat'}
        )

        self.betfair_stream.heartbeat()
        mock_send.assert_called_with(
                {'id': self.betfair_stream._unique_id, 'op': 'heartbeat'}
        )

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._send')
    def test_subscribe_to_markets(self, mock_send):
        market_filter = {'test': 123}
        market_data_filter = {'another_test': 123}
        self.betfair_stream.subscribe_to_markets(market_filter, market_data_filter, heartbeat_ms=1, conflate_ms=2,
                                                 segmentation_enabled=False)

        mock_send.assert_called_with(
                {'op': 'marketSubscription', 'marketFilter': market_filter, 'id': self.betfair_stream._unique_id,
                 'marketDataFilter': market_data_filter, 'initialClk': None, 'clk': None,
                 'heartbeatMs': 1, 'conflateMs': 2, 'segmentationEnabled': False}
        )
        self.mock_listener.register_stream.assert_called_with(self.betfair_stream._unique_id, 'marketSubscription')

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._send')
    def test_resubscribe_to_markets(self, mock_send):
        market_filter = {'test': 123}
        market_data_filter = {'another_test': 123}
        initial_clk = 'abcdef'
        clk = 'abc'
        self.betfair_stream.subscribe_to_markets(market_filter, market_data_filter, initial_clk, clk, heartbeat_ms=1,
                                                 conflate_ms=2, segmentation_enabled=False)

        mock_send.assert_called_with(
                {'op': 'marketSubscription', 'marketFilter': market_filter, 'id': self.betfair_stream._unique_id,
                 'marketDataFilter': market_data_filter, 'initialClk': initial_clk, 'clk': clk,
                 'heartbeatMs': 1, 'conflateMs': 2, 'segmentationEnabled': False}
        )
        assert not self.mock_listener.register_stream.called

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._send')
    def test_subscribe_to_orders(self, mock_send):
        initial_clk = 'abcdef'
        clk = 'abc'
        self.betfair_stream.subscribe_to_orders(initial_clk, clk, heartbeat_ms=1, conflate_ms=2,
                                                segmentation_enabled=False)
        mock_send.assert_called_with({
            'orderFilter': 'abcdef', 'id': self.betfair_stream._unique_id, 'op': 'orderSubscription', 'initialClk': 'abc', 'clk': None,
            'heartbeatMs': 1, 'conflateMs': 2, 'segmentationEnabled': False
        })
        self.mock_listener.register_stream.assert_called_with(self.betfair_stream._unique_id, 'orderSubscription')

    @mock.patch('ssl.wrap_socket')
    @mock.patch('socket.socket')
    def test_create_socket(self, mock_socket, mock_wrap_socket):
        self.betfair_stream._create_socket()

        mock_socket.assert_called_with(socket.AF_INET, socket.SOCK_STREAM)
        assert mock_wrap_socket.call_count == 1

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._data', return_value=False)
    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._receive_all', return_value='{}\r\n')
    def test_read_loop(self, mock_receive_all, mock_data):
        mock_socket = mock.Mock()
        self.betfair_stream._socket = mock_socket

        self.betfair_stream._running = True
        threading.Thread(target=self.betfair_stream._read_loop).start()

        for i in range(0, 2):
            time.sleep(0.1)
        self.betfair_stream._running = False
        time.sleep(0.1)

        mock_data.assert_called_with('{}')
        mock_receive_all.assert_called_with()
        assert self.betfair_stream.datetime_last_received is not None
        assert self.betfair_stream.receive_count > 0

    def test_receive_all(self):
        mock_socket = mock.Mock()
        data_return_value = b'{"op":"status"}\r\n'
        mock_socket.recv.return_value = data_return_value
        self.betfair_stream._socket = mock_socket

        data = self.betfair_stream._receive_all()
        assert data == ''

        self.betfair_stream._running = True
        data = self.betfair_stream._receive_all()
        mock_socket.recv.assert_called_with(self.buffer_size)
        assert data == data_return_value.decode('utf-8')

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream.stop')
    def test_receive_all_closed(self, mock_stop):
        mock_socket = mock.Mock()
        data_return_value = b''
        mock_socket.recv.return_value = data_return_value
        self.betfair_stream._socket = mock_socket
        self.betfair_stream._running = True

        with self.assertRaises(SocketError):
            self.betfair_stream._receive_all()
        mock_stop.assert_called_with()

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream.stop')
    def test_receive_all_error(self, mock_stop):
        mock_socket = mock.Mock()
        self.betfair_stream._socket = mock_socket

        self.betfair_stream._running = True
        mock_socket.recv.side_effect = socket.error()
        with self.assertRaises(SocketError):
            self.betfair_stream._receive_all()
        mock_stop.assert_called_with()

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream.stop')
    def test_receive_all_timeout(self, mock_stop):
        mock_socket = mock.Mock()
        self.betfair_stream._socket = mock_socket

        self.betfair_stream._running = True
        mock_socket.recv.side_effect = socket.timeout()
        with self.assertRaises(SocketError):
            self.betfair_stream._receive_all()
        mock_stop.assert_called_with()

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream.stop')
    def test_data(self, mock_stop):
        received_data = {"op": "status"}
        with self.assertRaises(ListenerError):
            self.betfair_stream._data(received_data)

        self.mock_listener.on_data.assert_called_with(received_data)
        assert mock_stop.called

        self.mock_listener.on_data.return_value = True
        self.betfair_stream._data(received_data)

        self.mock_listener.on_data.assert_called_with(received_data)
        assert mock_stop.call_count == 1

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream.authenticate')
    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream._connect')
    def test_send(self, mock_connect, mock_authenticate):
        mock_socket = mock.Mock()
        self.betfair_stream._socket = mock_socket
        message = {'message': 1}

        self.betfair_stream._send(message)
        assert mock_connect.call_count == 1
        assert mock_authenticate.call_count == 1
        assert mock_socket.send.call_count == 1

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream.stop')
    def test_send_timeout(self, mock_stop):
        self.betfair_stream._running = True
        mock_socket = mock.Mock()
        self.betfair_stream._socket = mock_socket
        mock_socket.send.side_effect = socket.timeout()
        message = {'message': 1}

        with self.assertRaises(SocketError):
            self.betfair_stream._send(message)
        mock_stop.assert_called_with()

    @mock.patch('betfairlightweight.streaming.betfairstream.BetfairStream.stop')
    def test_send_error(self, mock_stop):
        self.betfair_stream._running = True
        mock_socket = mock.Mock()
        self.betfair_stream._socket = mock_socket
        mock_socket.send.side_effect = socket.error()
        message = {'message': 1}

        with self.assertRaises(SocketError):
            self.betfair_stream._send(message)
        mock_stop.assert_called_with()

    def test_repr(self):
        assert repr(self.betfair_stream) == '<BetfairStream>'

    def test_str(self):
        assert str(self.betfair_stream) == '<BetfairStream [not running]>'
        self.betfair_stream._running = True
        assert str(self.betfair_stream) == '<BetfairStream [running]>'


class HistoricalStreamTest(unittest.TestCase):

    def setUp(self):
        self.directory = 'tests/resources/historicaldata/BASIC-1.132153978'
        self.listener = mock.Mock()
        self.stream = HistoricalStream(self.directory, self.listener)

    def test_init(self):
        assert self.stream.directory == self.directory
        assert self.stream.listener == self.listener
        assert self.stream._running is False

    @mock.patch('betfairlightweight.endpoints.streaming.HistoricalStream._read_loop')
    def test_start(self, mock_read_loop):
        self.stream.start()
        mock_read_loop.assert_called_with()
        assert self.stream._running is True

    @mock.patch('betfairlightweight.streaming.betfairstream.HistoricalStream._read_loop')
    @mock.patch('betfairlightweight.streaming.betfairstream.threading')
    def test_start_thread(self, mock_threading, mock_read_loop):
        self.stream.start(async_=True)
        mock_threading.Thread.assert_called_with(name='HistoricalStream', target=mock_read_loop)
        assert self.stream._running is True

    def test_stop(self):
        self.stream._running = True
        self.stream.stop()
        assert self.stream._running is False

    @mock.patch('betfairlightweight.streaming.betfairstream.HistoricalStream.stop')
    def test__read_loop(self, mock_stop):
        self.stream._running = True
        self.stream._read_loop()
        self.assertEqual(self.listener.on_data.call_count, 480)
        self.listener.on_data.snap()
        mock_stop.assert_called_with()
        self.assertTrue(self.stream._running)


class HistoricalGeneratorStreamTest(unittest.TestCase):

    def setUp(self):
        self.directory = 'tests/resources/historicaldata/BASIC-1.132153978'
        self.listener = mock.Mock()
        self.stream = HistoricalGeneratorStream(self.directory, self.listener)

    def test_init(self):
        assert self.stream.directory == self.directory
        assert self.stream.listener == self.listener
        assert self.stream._running is False

    @mock.patch('betfairlightweight.streaming.betfairstream.HistoricalGeneratorStream._read_loop')
    def test_get_generator(self, mock_read_loop):
        self.assertEqual(self.stream.get_generator(), mock_read_loop)

    @mock.patch('betfairlightweight.streaming.betfairstream.HistoricalGeneratorStream.stop')
    def test__read_loop(self, mock_stop):
        data = [i for i in self.stream._read_loop()]
        self.assertEqual(len(data), 480)
        self.assertEqual(self.listener.on_data.call_count, 480)
        self.listener.on_data.snap()
        mock_stop.assert_called_with()
        self.assertTrue(self.stream._running)



import unittest

import betfairlightweight
from betfairlightweight import StreamListener


class HistoricalStreamTest(unittest.TestCase):

    def test_historical_stream(self):
        trading = betfairlightweight.APIClient('username', 'password', app_key='appKey')
        stream = trading.streaming.create_historical_stream(
            directory='tests/resources/historicaldata/BASIC-1.132153978',
            listener=StreamListener()
        )
        stream.start(async_=False)

        assert stream.listener.stream_type == 'marketSubscription'
        assert stream.listener.stream_unique_id == 'HISTORICAL'
        assert stream.listener.clk == '3522512789'

        assert stream.listener.stream._updates_processed == 480
        assert len(stream.listener.stream._caches) == 1

        market = stream.listener.stream._caches.get('1.132153978')
        assert len(market.runners) == 14
        assert stream._running is False



from .utils import to_camel_case


def streaming_market_filter(market_ids=None, bsp_market=None, betting_types=None, event_type_ids=None, event_ids=None,
                            turn_in_play_enabled=None, market_types=None, venues=None, country_codes=None,
                            race_types=None):
    """
    :param list market_ids: filter market data to data pertaining to specific marketIds.
    :param list event_type_ids: filter market data to data pertaining to specific event_type ids.
    :param list event_ids: filter market data to data pertaining to specific event ids.
    :param bool turn_in_play_enabled: restriction on whether market will turn in play or not, not supplied returns all.
    :param list venues: restrict markets by venue (only horse racing has venue at the moment)
    :param bool bsp_market: restriction on bsp, not supplied will return all.
    :param list betting_types: filter on betting types
    :param list market_types: filter market data by market types.
    :param list country_codes: filter based on country codes
    :param list race_types: filter race types

    :return: dict
    """
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def streaming_market_data_filter(fields=None, ladder_levels=None):
    """
    :param list fields: EX_BEST_OFFERS_DISP, EX_BEST_OFFERS, EX_ALL_OFFERS, EX_TRADED,
    EX_TRADED_VOL, EX_LTP, EX_MARKET_DEF, SP_TRADED, SP_PROJECTED
    :param int ladder_levels: 1->10

    :return: dict
    """
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def streaming_order_filter(include_overall_position=None, customer_strategy_refs=None,
                           partition_matched_by_strategy_ref=None):
    """
    :param bool include_overall_position: Returns overall / net position (OrderRunnerChange.mb / OrderRunnerChange.ml)
    :param list customer_strategy_refs: Restricts to specified customerStrategyRefs; this will filter orders and
    StrategyMatchChanges accordingly (Note: overall postition is not filtered)
    :param bool partition_matched_by_strategy_ref: Returns strategy positions (OrderRunnerChange.smc=
    Map<customerStrategyRef, StrategyMatchChange>) - these are sent in delta format as per overall position.

    :return: dict
    """
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def time_range(from_=None, to=None):  # todo datetime conversion
    """
    :param str from_:
    :param str to:

    :return: dict
    """
    args = locals().copy()
    return {
        k.replace('_', ''): v for k, v in args.items()
    }


def market_filter(text_query=None, event_type_ids=None, event_ids=None, competition_ids=None, market_ids=None,
                  venues=None, bsp_only=None, turn_in_play_enabled=None, in_play_only=None, market_betting_types=None,
                  market_countries=None, market_type_codes=None, market_start_time=None, with_orders=None,
                  race_types=None):
    """
    :param str text_query: restrict markets by text associated with it, e.g name, event, comp.
    :param list event_type_ids: filter market data to data pertaining to specific event_type ids.
    :param list event_ids: filter market data to data pertaining to specific event ids.
    :param list competition_ids: filter market data to data pertaining to specific competition ids.
    :param list market_ids: filter market data to data pertaining to specific marketIds.
    :param list venues: restrict markets by venue (only horse racing has venue at the moment)
    :param bool bsp_only: restriction on bsp, not supplied will return all.
    :param bool turn_in_play_enabled: restriction on whether market will turn in play or not, not supplied returns all.
    :param bool in_play_only: restriction to currently inplay, not supplied returns all.
    :param list market_betting_types: filter market data by market betting types.
    :param list market_countries: filter market data by country codes.
    :param list market_type_codes: filter market data to match the type of market e.g. MATCH_ODDS.
    :param dict market_start_time: filter market data by time at which it starts.
    :param str with_orders: filter market data by specified order status.
    :param list race_types: filter race types.

    :return: dict
    """
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def price_data(sp_available=False, sp_traded=False, ex_best_offers=False, ex_all_offers=False, ex_traded=False):
    """
    Create PriceData filter list from all args passed as True.
    :param bool sp_available: Amount available for the BSP auction.
    :param bool sp_traded: Amount traded in the BSP auction.
    :param bool ex_best_offers: Only the best prices available for each runner, to requested price depth.
    :param bool ex_all_offers: trumps EX_BEST_OFFERS if both settings are present
    :param bool ex_traded: Amount traded on the exchange.

    :returns: string values of all args specified as True.
    :rtype: list
    """
    args = locals().copy()
    return [
        k.upper() for k, v in args.items() if v is True
    ]


def ex_best_offers_overrides(best_prices_depth=None, rollup_model=None, rollup_limit=None,
                             rollup_liability_threshold=None, rollup_liability_factor=None):
    """
    Create filter to specify whether to accumulate market volume info, how deep a book to return and rollup methods if
    accumulation is selected.
    :param int best_prices_depth: The maximum number of prices to return on each side for each runner.
    :param str rollup_model: method to use to accumulate market orders.
    :param int rollup_limit: The volume limit to use when rolling up returned sizes. The exact definition of the limit
    depends on the rollupModel.
                        If no limit is provided it will use minimum stake
    :param float rollup_liability_threshold: Only applicable when rollupModel is MANAGED_LIABILITY. The rollup model
    switches from being stake based to liability based at the smallest lay price which is >= rollupLiabilityThreshold
    :param int rollup_liability_factor: Only applicable when rollupModel is MANAGED_LIABILITY. (rollupLiabilityFactor *
    rollupLimit) is the minimum liabilty the user is deemed to be comfortable with. After the rollupLiabilityThreshold
    price subsequent volumes will be rolled up to minimum value such that the liability >= the minimum liability.

    :returns: parameters for inclusion in market data requests.
    :rtype: dict
    """

    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def price_projection(price_data=None, ex_best_offers_overrides=None, virtualise=True,
                     rollover_stakes=False):
    """
    Selection criteria of the returning price data.
    :param list price_data: PriceData filter to specify what market data we wish to receive.
    :param dict ex_best_offers_overrides: define order book depth, rollup method.
    :param bool virtualise: whether to receive virtualised prices also.
    :param bool rollover_stakes: whether to accumulate volume at each price as sum of volume at that price and all better
    prices.

    :returns: price data criteria for market data.
    :rtype: dict
    """
    if price_data is None:
        price_data = []
    if ex_best_offers_overrides is None:
        ex_best_offers_overrides = {}
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def place_instruction(order_type, selection_id, side, handicap=None, limit_order=None, limit_on_close_order=None,
                      market_on_close_order=None, customer_order_ref=None):
    """
    Create order instructions to place an order at exchange.
    :param str order_type: define type of order to place.
    :param int selection_id: selection on which to place order
    :param float handicap: handicap if placing order on asianhandicap type market
    :param str side: side of order
    :param resources.LimitOrder limit_order: if orderType is a limitOrder structure details of the order.
    :param resources.LimitOnCloseOrder limit_on_close_order: if orderType is a
    limitOnCloseOrder structure details of the order.
    :param resources.MarketOnCloseOrder market_on_close_order: if orderType is
    a marketOnCloseOrder structure details of the order.
    :param str customer_order_ref: an optional reference customers can set to identify instructions..

    :return: orders to place.
    :rtype: dict
    """

    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def limit_order(price, persistence_type=None, size=None, time_in_force=None, min_fill_size=None, bet_target_type=None,
                bet_target_size=None):
    """
    Create a limit order to send to exchange.
    :param float size: amount in account currency to be sent.
    :param float price: price at which the order is to be sent.
    :param str persistence_type: what happens to order at turn in play.
    :param str time_in_force: specify if it is FillOrKill/FillAndKill. This value takes precedence over any
    PersistenceType value chosen.
    :param float min_fill_size: the minimum amount to be filled for FillAndKill.
    :param str bet_target_type: Specify the type of Target, bet to certain backer profit or certain payout value.
                          Used to adjust to lower stakes if filled at better levels.
    :param float bet_target_size: Size of payout of profit to bet.

    :returns: Order information to place a limit order.
    :rtype: dict
    """
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def limit_on_close_order(liability, price):
    """
    Create limit order for the closing auction.
    :param float liability: amount to bet.
    :param float price: price at which to bet

    :returns: Order information to place a limit on close order.
    :rtype: dict
    """
    return locals().copy()


def market_on_close_order(liability):
    """
    Create market order to be placed in the closing auction.
    :param float liability: amount to bet.

    :returns: Order information to place a market on close order.
    :rtype: dict
    """
    return locals().copy()


def cancel_instruction(bet_id, size_reduction=None):
    """
    Instruction to fully or partially cancel an order (only applies to LIMIT orders)
    :param str bet_id: identifier of the bet to cancel.
    :param float size_reduction: If supplied then this is a partial cancel.

    :returns: cancellation report detailing status, cancellation requested and actual cancellation details.
    :rtype: dict
    """
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def replace_instruction(bet_id, new_price):
    """
    Instruction to replace a LIMIT or LIMIT_ON_CLOSE order at a new price.
    Original order will be cancelled and a new order placed at the new price for the remaining stake.
    :param str bet_id: Unique identifier for the bet
    :param float new_price: The price to replace the bet at

    :returns: replace report detailing status, replace requested and actual replace details.
    :rtype: dict
    """
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }


def update_instruction(bet_id, new_persistence_type):
    """
    Instruction to update LIMIT bet's persistence of an order that do not affect exposure
    :param str bet_id: Unique identifier for the bet
    :param str new_persistence_type: The new persistence type to update this bet to.

    :returns: update report detailing status, update requested and update details.
    :rtype: dict
    """
    args = locals().copy()
    return {
        to_camel_case(k): v for k, v in args.items() if v is not None
    }



from .baseresource import BaseResource


class EventType(object):
    """
    :type id: unicode
    :type name: unicode
    """

    def __init__(self, id, name):
        self.id = id
        self.name = name


class EventTypeResult(BaseResource):
    """
    :type event_type: EventType
    :type market_count: int
    """

    def __init__(self, **kwargs):
        super(EventTypeResult, self).__init__(**kwargs)
        self.market_count = kwargs.get('marketCount')
        self.event_type = EventType(**kwargs.get('eventType'))


class Competition(object):
    """
    :type id: unicode
    :type name: unicode
    """

    def __init__(self, id, name):
        self.id = id
        self.name = name


class CompetitionResult(BaseResource):
    """
    :type competition: Competition
    :type competition_region: unicode
    :type market_count: int
    """

    def __init__(self, **kwargs):
        super(CompetitionResult, self).__init__(**kwargs)
        self.market_count = kwargs.get('marketCount')
        self.competition_region = kwargs.get('competitionRegion')
        self.competition = Competition(**kwargs.get('competition'))


class TimeRange(object):
    """
    :type _from: datetime.datetime
    :type to: datetime.datetime
    """

    def __init__(self, **kwargs):
        self._from = BaseResource.strip_datetime(kwargs.get('from'))
        self.to = BaseResource.strip_datetime(kwargs.get('to'))


class TimeRangeResult(BaseResource):
    """
    :type market_count: int
    :type time_range: TimeRange
    """

    def __init__(self, **kwargs):
        super(TimeRangeResult, self).__init__(**kwargs)
        self.market_count = kwargs.get('marketCount')
        self.time_range = TimeRange(**kwargs.get('timeRange'))


class Event(object):
    """
    :type country_code: unicode
    :type id: unicode
    :type name: unicode
    :type open_date: datetime.datetime
    :type time_zone: unicode
    :type venue: unicode
    """

    def __init__(self, id, openDate, timezone, name, countryCode=None, venue=None):
        self.id = id
        self.open_date = BaseResource.strip_datetime(openDate)
        self.time_zone = timezone
        self.country_code = countryCode
        self.name = name
        self.venue = venue


class EventResult(BaseResource):
    """
    :type event: Event
    :type market_count: int
    """

    def __init__(self, **kwargs):
        super(EventResult, self).__init__(**kwargs)
        self.market_count = kwargs.get('marketCount')
        self.event = Event(**kwargs.get('event'))


class MarketTypeResult(BaseResource):
    """
    :type market_count: int
    :type market_type: unicode
    """

    def __init__(self, **kwargs):
        super(MarketTypeResult, self).__init__(**kwargs)
        self.market_count = kwargs.get('marketCount')
        self.market_type = kwargs.get('marketType')


class CountryResult(BaseResource):
    """
    :type country_code: unicode
    :type market_count: int
    """

    def __init__(self, **kwargs):
        super(CountryResult, self).__init__(**kwargs)
        self.market_count = kwargs.get('marketCount')
        self.country_code = kwargs.get('countryCode')


class VenueResult(BaseResource):
    """
    :type market_count: int
    :type venue: unicode
    """

    def __init__(self, **kwargs):
        super(VenueResult, self).__init__(**kwargs)
        self.market_count = kwargs.get('marketCount')
        self.venue = kwargs.get('venue')


class LineRangeInfo(object):
    """
    :type marketUnit: unicode
    :type interval: float
    :type minUnitValue: float
    :type maxUnitValue: float
    """
    def __init__(self, marketUnit, interval, minUnitValue, maxUnitValue):
        self.market_unit = marketUnit
        self.interval = interval
        self.min_unit_value = minUnitValue
        self.max_unit_value = maxUnitValue


class PriceLadderDescription(object):
    """
    :type type: unicode
    """

    def __init__(self, type):
        self.type = type

    def serialise(self):
        return {'type': self.type}


class MarketCatalogueDescription(object):
    """
    :type betting_type: unicode
    :type bsp_market: bool
    :type clarifications: unicode
    :type discount_allowed: bool
    :type each_way_divisor: float
    :type market_base_rate: float
    :type market_time: datetime.datetime
    :type market_type: unicode
    :type persistence_enabled: bool
    :type regulator: unicode
    :type rules: unicode
    :type rules_has_date: bool
    :type suspend_time: datetime.datetime
    :type turn_in_play_enabled: bool
    :type wallet: unicode
    """

    def __init__(self, bettingType, bspMarket, marketTime, marketType, suspendTime, turnInPlayEnabled,
                 wallet=None, discountAllowed=None, marketBaseRate=None, persistenceEnabled=None, regulator=None,
                 rules=None, rulesHasDate=None, eachWayDivisor=None, clarifications=None, priceLadderDescription=None,
                 keyLineDefinition=None, lineRangeInfo=None, raceType=None):
        self.betting_type = bettingType
        self.bsp_market = bspMarket
        self.discount_allowed = discountAllowed
        self.market_base_rate = marketBaseRate
        self.market_time = BaseResource.strip_datetime(marketTime)
        self.market_type = marketType
        self.persistence_enabled = persistenceEnabled
        self.regulator = regulator
        self.rules = rules
        self.rules_has_date = rulesHasDate
        self.suspend_time = BaseResource.strip_datetime(suspendTime)
        self.turn_in_play_enabled = turnInPlayEnabled
        self.wallet = wallet
        self.each_way_divisor = eachWayDivisor
        self.clarifications = clarifications
        self.price_ladder_description = PriceLadderDescription(**priceLadderDescription) if priceLadderDescription else None
        self.line_range_info = LineRangeInfo(**lineRangeInfo) if lineRangeInfo else None
        self.race_type = raceType


class RunnerCatalogue(object):
    """
    :type handicap: float
    :type metadata: dict
    :type runner_name: unicode
    :type selection_id: int
    :type sort_priority: int
    """

    def __init__(self, **kwargs):
        self.selection_id = kwargs.get('selectionId')
        self.runner_name = kwargs.get('runnerName')
        self.sort_priority = kwargs.get('sortPriority')
        self.handicap = kwargs.get('handicap')
        self.metadata = kwargs.get('metadata')

    def __str__(self):
        return 'RunnerCatalogue: %s' % self.selection_id

    def __repr__(self):
        return '<RunnerCatalogue>'


class MarketCatalogue(BaseResource):
    """
    :type competition: Competition
    :type description: MarketCatalogueDescription
    :type event: Event
    :type event_type: EventType
    :type market_id: unicode
    :type market_name: unicode
    :type market_start_time: datetime.datetime
    :type runners: list[RunnerCatalogue]
    :type total_matched: float
    """

    def __init__(self, **kwargs):
        super(MarketCatalogue, self).__init__(**kwargs)
        self.market_id = kwargs.get('marketId')
        self.market_name = kwargs.get('marketName')
        self.total_matched = kwargs.get('totalMatched')
        self.market_start_time = self.strip_datetime(kwargs.get('marketStartTime'))
        self.competition = Competition(**kwargs.get('competition')) if kwargs.get('competition') else None
        self.event = Event(**kwargs.get('event')) if kwargs.get('event') else None
        self.event_type = EventType(**kwargs.get('eventType')) if kwargs.get('eventType') else None
        self.description = MarketCatalogueDescription(**kwargs.get('description')) if \
            kwargs.get('description') else None
        self.runners = [RunnerCatalogue(**i) for i in kwargs.get('runners', [])]


"""
__slots__ is a terrible hack with nasty, hard-to-fathom side
effects that should only be used by programmers at grandmaster and
wizard levels. Unfortunately it has gained an enormous undeserved
popularity amongst the novices and apprentices, who should know
better than to use this magic incantation casually.
"""


class Slotable(object):
    __slots__ = []

    def __getstate__(self):
        return {slot: getattr(self, slot) for slot in self.__slots__}

    def __setstate__(self, d):
        for slot in d:
            setattr(self, slot, d[slot])


class PriceSize(Slotable):
    """
    :type price: float
    :type size: float
    """

    __slots__ = [
        'price', 'size'
    ]

    def __init__(self, price, size):
        self.price = price
        self.size = size

    def __str__(self):
        return 'Price: %s Size: %s' % (self.price, self.size)


class RunnerBookSP(object):
    """
    :type actual_sp: float
    :type back_stake_taken: list[PriceSize]
    :type far_price: float
    :type lay_liability_taken: list[PriceSize]
    :type near_price: float
    """

    def __init__(self, nearPrice=None, farPrice=None, backStakeTaken=None, layLiabilityTaken=None, actualSP=None):
        self.near_price = nearPrice
        self.far_price = farPrice
        self.actual_sp = actualSP
        self.back_stake_taken = [PriceSize(**i) for i in backStakeTaken]
        self.lay_liability_taken = [PriceSize(**i) for i in layLiabilityTaken]


class RunnerBookEX(object):
    """
    :type available_to_back: list[PriceSize]
    :type available_to_lay: list[PriceSize]
    :type traded_volume: list[PriceSize]
    """

    def __init__(self, availableToBack=None, availableToLay=None, tradedVolume=None):
        self.available_to_back = [PriceSize(**i) for i in availableToBack]
        self.available_to_lay = [PriceSize(**i) for i in availableToLay]
        self.traded_volume = [PriceSize(**i) for i in tradedVolume]


class RunnerBookOrder(object):
    """
    :type avg_price_matched: float
    :type bet_id: unicode
    :type bsp_liability: float
    :type order_type: unicode
    :type persistence_type: unicode
    :type placed_date: datetime.datetime
    :type price: float
    :type side: unicode
    :type size: float
    :type size_cancelled: float
    :type size_lapsed: float
    :type size_matched: float
    :type size_remaining: float
    :type size_voided: float
    :type status: unicode
    """

    def __init__(self, betId, avgPriceMatched, bspLiability, orderType, persistenceType, placedDate, price, side,
                 sizeCancelled, sizeLapsed, sizeMatched, sizeRemaining, sizeVoided, status, size,
                 customerStrategyRef=None, customerOrderRef=None):
        self.bet_id = betId
        self.avg_price_matched = avgPriceMatched
        self.bsp_liability = bspLiability
        self.order_type = orderType
        self.persistence_type = persistenceType
        self.placed_date = BaseResource.strip_datetime(placedDate)
        self.price = price
        self.side = side
        self.size_cancelled = sizeCancelled
        self.size_lapsed = sizeLapsed
        self.size_matched = sizeMatched
        self.size_remaining = sizeRemaining
        self.size_voided = sizeVoided
        self.status = status
        self.size = size
        self.customer_strategy_ref = customerStrategyRef
        self.customer_order_ref = customerOrderRef


class RunnerBookMatch(object):
    """
    :type bet_id: unicode
    :type match_date: datetime.datetime
    :type match_id: unicode
    :type price: float
    :type side: unicode
    :type size: float
    """

    def __init__(self, price, side, size, betId=None, matchId=None, matchDate=None):
        self.bet_id = betId
        self.match_id = matchId
        self.price = price
        self.side = side
        self.size = size
        self.match_date = BaseResource.strip_datetime(matchDate)


class RunnerBook(object):
    """
    :type adjustment_factor: float
    :type ex: RunnerBookEX
    :type handicap: float
    :type last_price_traded: float
    :type matches: list[RunnerBookMatch]
    :type orders: list[RunnerBookOrder]
    :type removal_date: datetime.datetime
    :type selection_id: int
    :type sp: RunnerBookSP
    :type status: unicode
    :type total_matched: float
    """

    def __init__(self, selectionId, status, handicap, adjustmentFactor=None, lastPriceTraded=None, totalMatched=None,
                 removalDate=None, sp=None, ex=None, orders=None, matches=None, matchesByStrategy=None):
        self.selection_id = selectionId
        self.status = status
        self.total_matched = totalMatched
        self.adjustment_factor = adjustmentFactor
        self.handicap = handicap
        self.last_price_traded = lastPriceTraded
        self.removal_date = BaseResource.strip_datetime(removalDate)
        self.sp = RunnerBookSP(**sp) if sp else None
        self.ex = RunnerBookEX(**ex) if ex else None
        self.orders = [RunnerBookOrder(**i) for i in orders] if orders else []
        self.matches = [RunnerBookMatch(**i) for i in matches] if matches else []
        self.matches_by_strategy = matchesByStrategy

    def __str__(self):
        return 'RunnerBook: %s' % self.selection_id

    def __repr__(self):
        return '<RunnerBook>'


class KeyLine(object):

    def __init__(self, **kwargs):
        if 'keyLine' in kwargs:
            self.key_line = [KeyLineSelection(**i) for i in kwargs['keyLine']]
        elif 'kl' in kwargs:
            self.key_line = [KeyLineSelection(**i) for i in kwargs['kl']]


class KeyLineSelection(object):
    """
    :type selectionId: int
    :type handicap: float
    """

    def __init__(self, **kwargs):
        if 'selectionId' in kwargs:
            self.selection_id = kwargs['selectionId']
        elif 'id' in kwargs:
            self.selection_id = kwargs['id']

        if 'handicap' in kwargs:
            self.handicap = kwargs['handicap']
        elif 'hc' in kwargs:
            self.handicap = kwargs['hc']


class MarketBook(BaseResource):
    """
    :type bet_delay: int
    :type bsp_reconciled: bool
    :type complete: bool
    :type cross_matching: bool
    :type inplay: bool
    :type is_market_data_delayed: bool
    :type last_match_time: datetime.datetime
    :type market_id: unicode
    :type number_of_active_runners: int
    :type number_of_runners: int
    :type number_of_winners: int
    :type publish_time: datetime.datetime
    :type runners: list[RunnerBook]
    :type runners_voidable: bool
    :type status: unicode
    :type total_available: float
    :type total_matched: float
    :type version: int
    """

    def __init__(self, **kwargs):
        self.streaming_unique_id = kwargs.pop('streaming_unique_id', None)
        self.streaming_update = kwargs.pop('streaming_update', None)
        self.market_definition = kwargs.pop('market_definition', None)
        super(MarketBook, self).__init__(**kwargs)
        self.market_id = kwargs.get('marketId')
        self.bet_delay = kwargs.get('betDelay')
        self.bsp_reconciled = kwargs.get('bspReconciled')
        self.complete = kwargs.get('complete')
        self.cross_matching = kwargs.get('crossMatching')
        self.inplay = kwargs.get('inplay')
        self.is_market_data_delayed = kwargs.get('isMarketDataDelayed')
        self.last_match_time = self.strip_datetime(kwargs.get('lastMatchTime'))
        self.number_of_active_runners = kwargs.get('numberOfActiveRunners')
        self.number_of_runners = kwargs.get('numberOfRunners')
        self.number_of_winners = kwargs.get('numberOfWinners')
        self.runners_voidable = kwargs.get('runnersVoidable')
        self.status = kwargs.get('status')
        self.total_available = kwargs.get('totalAvailable')
        self.total_matched = kwargs.get('totalMatched')
        self.version = kwargs.get('version')
        self.runners = [RunnerBook(**i) for i in kwargs.get('runners')]
        self.publish_time = self.strip_datetime(kwargs.get('publishTime'))
        self.publish_time_epoch = kwargs.get('publishTime')
        self.key_line_description = KeyLine(**kwargs.get('keyLineDescription')
                                            ) if kwargs.get('keyLineDescription') else None
        self.price_ladder_definition = PriceLadderDescription(**kwargs.get('priceLadderDefinition')
                                                              ) if kwargs.get('priceLadderDefinition') else None


class CurrentOrder(object):
    """
    :type average_price_matched: float
    :type bet_id: unicode
    :type bsp_liability: float
    :type customer_order_ref: unicode
    :type customer_strategy_ref: unicode
    :type handicap: float
    :type market_id: unicode
    :type matched_date: datetime.datetime
    :type order_type: unicode
    :type persistence_type: unicode
    :type placed_date: datetime.datetime
    :type price_size: PriceSize
    :type regulator_code: unicode
    :type selection_id: int
    :type side: unicode
    :type size_cancelled: float
    :type size_lapsed: float
    :type size_matched: float
    :type size_remaining: float
    :type size_voided: float
    :type status: unicode
    """

    def __init__(self, betId, averagePriceMatched, bspLiability, handicap, marketId, orderType, persistenceType,
                 placedDate, regulatorCode, selectionId, side, sizeCancelled, sizeLapsed, sizeMatched, sizeRemaining,
                 sizeVoided, status, priceSize, matchedDate=None, customerStrategyRef=None, customerOrderRef=None):
        self.bet_id = betId
        self.average_price_matched = averagePriceMatched
        self.bsp_liability = bspLiability
        self.handicap = handicap
        self.market_id = marketId
        self.matched_date = BaseResource.strip_datetime(matchedDate)
        self.order_type = orderType
        self.persistence_type = persistenceType
        self.placed_date = BaseResource.strip_datetime(placedDate)
        self.regulator_code = regulatorCode
        self.selection_id = selectionId
        self.side = side
        self.size_cancelled = sizeCancelled
        self.size_lapsed = sizeLapsed
        self.size_matched = sizeMatched
        self.size_remaining = sizeRemaining
        self.size_voided = sizeVoided
        self.status = status
        self.customer_strategy_ref = customerStrategyRef
        self.customer_order_ref = customerOrderRef
        self.price_size = PriceSize(**priceSize)


class CurrentOrders(BaseResource):
    """
    :type more_available: bool
    :type orders: list[CurrentOrder]
    """

    def __init__(self, **kwargs):
        self.streaming_unique_id = kwargs.pop('streaming_unique_id', None)
        self.streaming_update = kwargs.pop('streaming_update', None)
        self.publish_time = kwargs.pop('publish_time', None)
        super(CurrentOrders, self).__init__(**kwargs)
        self.more_available = kwargs.get('moreAvailable')
        self.orders = [CurrentOrder(**i) for i in kwargs.get('currentOrders')]


class ItemDescription(object):
    """
    :type event_desc: unicode
    :type event_type_desc: unicode
    :type market_desc: unicode
    :type market_start_time: datetime
    :type market_type: unicode
    :type number_of_winners: int
    :type runner_desc: unicode
    :type each_way_divisor: unicode
    """

    def __init__(self, eventDesc=None, eventTypeDesc=None, marketDesc=None, marketStartTime=None, marketType=None,
                 numberOfWinners=None, runnerDesc=None, eachWayDivisor=None):
        self.event_desc = eventDesc
        self.event_type_desc = eventTypeDesc
        self.market_desc = marketDesc
        self.market_start_time = BaseResource.strip_datetime(marketStartTime)
        self.market_type = marketType
        self.number_of_winners = numberOfWinners
        self.runner_desc = runnerDesc
        self.each_way_divisor = eachWayDivisor


class ClearedOrder(object):
    """
    :type bet_count: int
    :type bet_id: unicode
    :type bet_outcome: unicode
    :type customer_order_ref: unicode
    :type customer_strategy_ref: unicode
    :type event_id: unicode
    :type event_type_id: unicode
    :type handicap: float
    :type last_matched_date: datetime.datetime
    :type market_id: unicode
    :type order_type: unicode
    :type persistence_type: unicode
    :type placed_date: datetime.datetime
    :type price_matched: float
    :type price_reduced: bool
    :type price_requested: float
    :type profit: float
    :type commission: float
    :type selection_id: int
    :type settled_date: datetime.datetime
    :type side: unicode
    :type size_settled: float
    :type size_cancelled float
    :type item_description ItemDescription
    """

    def __init__(self, **kwargs):
        self.bet_id = kwargs.get("betId")
        self.bet_count = kwargs.get("betCount")
        self.bet_outcome = kwargs.get("betOutcome")
        self.event_id = kwargs.get("eventId")
        self.event_type_id = kwargs.get("eventTypeId")
        self.handicap = kwargs.get("handicap")
        self.last_matched_date = BaseResource.strip_datetime(kwargs.get("lastMatchedDate"))
        self.market_id = kwargs.get("marketId")
        self.order_type = kwargs.get("orderType")
        self.persistence_type = kwargs.get("persistenceType")
        self.placed_date = BaseResource.strip_datetime(kwargs.get("placedDate"))
        self.price_matched = kwargs.get("priceMatched")
        self.price_reduced = kwargs.get("priceReduced")
        self.price_requested = kwargs.get("priceRequested")
        self.profit = kwargs.get("profit")
        self.commission = kwargs.get("commission")
        self.selection_id = kwargs.get("selectionId")
        self.settled_date = BaseResource.strip_datetime(kwargs.get("settledDate"))
        self.side = kwargs.get("side")
        self.size_settled = kwargs.get("sizeSettled")
        self.size_cancelled = kwargs.get("sizeCancelled")
        self.customer_strategy_ref = kwargs.get("customerStrategyRef")
        self.customer_order_ref = kwargs.get("customerOrderRef")
        self.item_description = ItemDescription(**kwargs.get("itemDescription")
                                                ) if 'itemDescription' in kwargs else None


class ClearedOrders(BaseResource):
    """
    :type more_available: bool
    :type orders: list[ClearedOrder]
    """

    def __init__(self, **kwargs):
        super(ClearedOrders, self).__init__(**kwargs)
        self.more_available = kwargs.get('moreAvailable')
        self.orders = [ClearedOrder(**i) for i in kwargs.get('clearedOrders')]


class ProfitAndLosses(object):
    """
    :type if_lose: float
    :type if_place: float
    :type if_win: float
    :type selection_id: int
    """

    def __init__(self, selectionId, ifWin=None, ifLose=None, ifPlace=None):
        self.selection_id = selectionId
        self.if_win = ifWin
        self.if_lose = ifLose
        self.if_place = ifPlace


class MarketProfitLoss(BaseResource):
    """
    :type commission_applied: float
    :type market_id: unicode
    :type profit_and_losses: list[ProfitAndLosses]
    """

    def __init__(self, **kwargs):
        super(MarketProfitLoss, self).__init__(**kwargs)
        self.market_id = kwargs.get('marketId')
        self.commission_applied = kwargs.get('commissionApplied')
        self.profit_and_losses = [ProfitAndLosses(**i) for i in kwargs.get('profitAndLosses')]


class LimitOrder(object):
    """
    :type bet_target_size: float
    :type bet_target_type: unicode
    :type min_fill_size: float
    :type persistence_type: unicode
    :type price: float
    :type size: float
    :type time_in_force: unicode
    """

    def __init__(self, price, size=None, persistenceType=None, timeInForce=None, minFillSize=None, betTargetType=None,
                 betTargetSize=None):
        self.persistence_type = persistenceType
        self.price = price
        self.size = size
        self.time_in_force = timeInForce
        self.min_fill_size = minFillSize
        self.bet_target_type = betTargetType
        self.bet_target_size = betTargetSize


class LimitOnCloseOrder(object):
    """
    :type liability: float
    :type price: float
    """

    def __init__(self, liability, price):
        self.liability = liability
        self.price = price


class MarketOnCloseOrder(object):
    """
    :type liability: float
    """

    def __init__(self, liability):
        self.liability = liability


class PlaceOrderInstruction(object):
    """
    :type customer_order_ref: unicode
    :type handicap: float
    :type order: LimitOrder
    :type order_type: unicode
    :type selection_id: int
    :type side: unicode
    """

    def __init__(self, selectionId, side, orderType, limitOrder=None, limitOnCloseOrder=None, marketOnCloseOrder=None,
                 handicap=None, customerOrderRef=None):
        self.selection_id = selectionId
        self.side = side
        self.order_type = orderType
        self.handicap = handicap
        self.customer_order_ref = customerOrderRef
        self.limit_order = LimitOrder(**limitOrder) if limitOrder else None
        self.limit_on_close_order = LimitOnCloseOrder(**limitOnCloseOrder) if limitOnCloseOrder else None
        self.market_on_close_order = MarketOnCloseOrder(**marketOnCloseOrder) if marketOnCloseOrder else None


class PlaceOrderInstructionReports(object):
    """
    :type average_price_matched: float
    :type bet_id: unicode
    :type error_code: str
    :type instruction: PlaceOrderInstruction
    :type order_status: unicode
    :type placed_date: datetime.datetime
    :type size_matched: float
    :type status: unicode
    """

    def __init__(self, status, instruction=None, orderStatus=None, betId=None, averagePriceMatched=None,
                 sizeMatched=None, placedDate=None, errorCode=None):
        self.status = status
        self.order_status = orderStatus
        self.bet_id = betId
        self.average_price_matched = averagePriceMatched
        self.size_matched = sizeMatched
        self.placed_date = BaseResource.strip_datetime(placedDate)
        self.instruction = PlaceOrderInstruction(**instruction) if instruction else None
        self.error_code = errorCode


class PlaceOrders(BaseResource):
    """
    :type customer_ref: unicode
    :type error_code: str
    :type market_id: unicode
    :type place_instruction_reports: list[PlaceOrderInstructionReports]
    :type status: unicode
    """

    def __init__(self, **kwargs):
        super(PlaceOrders, self).__init__(**kwargs)
        self.market_id = kwargs.get('marketId')
        self.status = kwargs.get('status')
        self.customer_ref = kwargs.get('customerRef')
        self.error_code = kwargs.get('errorCode')
        self.place_instruction_reports = [
            PlaceOrderInstructionReports(**i) for i in kwargs.get('instructionReports')
        ]


class CancelOrderInstruction(object):
    """
    :type bet_id: unicode
    :type size_reduction: float
    """

    def __init__(self, betId, sizeReduction=None):
        self.bet_id = betId
        self.size_reduction = sizeReduction


class CancelOrderInstructionReports(object):
    """
    :type cancelled_date: datetime.datetime
    :type error_code: str
    :type instruction: CancelOrderInstruction
    :type size_cancelled: float
    :type status: unicode
    """

    def __init__(self, status, instruction, sizeCancelled=None, cancelledDate=None, errorCode=None):
        self.status = status
        self.size_cancelled = sizeCancelled
        self.cancelled_date = BaseResource.strip_datetime(cancelledDate)
        self.instruction = CancelOrderInstruction(**instruction)
        self.error_code = errorCode


class CancelOrders(BaseResource):
    """
    :type cancel_instruction_reports: list[CancelOrderInstructionReports]
    :type customer_ref: unicode
    :type error_code: str
    :type market_id: unicode
    :type status: unicode
    """

    def __init__(self, **kwargs):
        super(CancelOrders, self).__init__(**kwargs)
        self.market_id = kwargs.get('marketId')
        self.status = kwargs.get('status')
        self.customer_ref = kwargs.get('customerRef')
        self.error_code = kwargs.get('errorCode')
        self.cancel_instruction_reports = [
            CancelOrderInstructionReports(**i) for i in kwargs.get('instructionReports')
        ]


class UpdateOrderInstruction(object):
    """
    :type bet_id: unicode
    :type new_persistence_type: unicode
    """

    def __init__(self, betId, newPersistenceType):
        self.bet_id = betId
        self.new_persistence_type = newPersistenceType


class UpdateOrderInstructionReports(object):
    """
    :type error_code: str
    :type instruction: UpdateOrderInstruction
    :type status: unicode
    """

    def __init__(self, status, instruction, errorCode=None):
        self.status = status
        self.instruction = UpdateOrderInstruction(**instruction)
        self.error_code = errorCode


class UpdateOrders(BaseResource):
    """
    :type customer_ref: unicode
    :type error_code: str
    :type market_id: unicode
    :type status: unicode
    :type update_instruction_reports: list[UpdateOrderInstructionReports]
    """

    def __init__(self, **kwargs):
        super(UpdateOrders, self).__init__(**kwargs)
        self.market_id = kwargs.get('marketId')
        self.status = kwargs.get('status')
        self.customer_ref = kwargs.get('customerRef')
        self.error_code = kwargs.get('errorCode')
        self.update_instruction_reports = [
            UpdateOrderInstructionReports(**i) for i in kwargs.get('instructionReports')
        ]


class ReplaceOrderInstructionReports(object):
    """
    :type cancel_instruction_reports: CancelOrderInstructionReports
    :type error_code: str
    :type place_instruction_reports: PlaceOrderInstructionReports
    :type status: unicode
    """

    def __init__(self, status, cancelInstructionReport, placeInstructionReport, errorCode=None):
        self.status = status
        self.cancel_instruction_reports = CancelOrderInstructionReports(**cancelInstructionReport)
        self.place_instruction_reports = PlaceOrderInstructionReports(**placeInstructionReport)
        self.error_code = errorCode


class ReplaceOrders(BaseResource):
    """
    :type customer_ref: unicode
    :type error_code: str
    :type market_id: unicode
    :type replace_instruction_reports: list[ReplaceOrderInstructionReports]
    :type status: unicode
    """

    def __init__(self, **kwargs):
        super(ReplaceOrders, self).__init__(**kwargs)
        self.market_id = kwargs.get('marketId')
        self.status = kwargs.get('status')
        self.customer_ref = kwargs.get('customerRef')
        self.error_code = kwargs.get('errorCode')
        self.replace_instruction_reports = [
            ReplaceOrderInstructionReports(**i) for i in kwargs.get('instructionReports')
        ]



from __future__ import print_function

import datetime
import os
import unittest

from betfairlightweight import resources
from tests.unit.tools import create_mock_json


class RaceCardResourcesTest(unittest.TestCase):
    DATE_TIME_SENT = datetime.datetime(2003, 8, 4, 12, 30, 45)

    def test_racecard(self):
        rootdir = 'tests/resources/racecards'
        racecards = []
        for _, _, filenames in os.walk(rootdir):
            for json_racecard in [fname for fname in filenames
                                  if fname.startswith('racecards')]:
                mock_response = create_mock_json(os.path.join(rootdir, json_racecard))
                racecard = mock_response.json().get('result')
                racecards.append(
                    resources.RaceCard(date_time_sent=self.DATE_TIME_SENT,
                                       **racecard))
        assert 0 < len(racecards)
        assert all(isinstance(racecard, resources.RaceCard)
                   for racecard in racecards)



from ..streaming import (
    BaseListener,
    StreamListener,
    BetfairStream,
    HistoricalStream,
    HistoricalGeneratorStream
)


class Streaming(object):
    """
    Streaming operations.
    """

    def __init__(self, parent):
        """
        :param parent: API client.
        """
        self.client = parent

    def create_stream(self, unique_id=0, listener=None, timeout=11, buffer_size=1024, description='BetfairSocket',
                      host=None):
        """
        Creates BetfairStream.

        :param dict unique_id: Id used to start unique id's of the stream (+1 before every request)
        :param resources.Listener listener:  Listener class to use
        :param float timeout: Socket timeout
        :param int buffer_size: Socket buffer size
        :param str description: Betfair stream description
        :param str host: Host endpoint (prod (default) or integration)

        :rtype: BetfairStream
        """
        listener = listener if listener else BaseListener()
        return BetfairStream(
            unique_id,
            listener,
            app_key=self.client.app_key,
            session_token=self.client.session_token,
            timeout=timeout,
            buffer_size=buffer_size,
            description=description,
            host=host,
        )

    @staticmethod
    def create_historical_stream(directory, listener=None):
        """
        Uses streaming listener/cache to parse betfair
        historical data:
            https://historicdata.betfair.com/#/home

        :param str directory: Directory of betfair data
        :param BaseListener listener: Listener object

        :rtype: HistoricalStream
        """
        listener = listener if listener else BaseListener()
        listener.register_stream('HISTORICAL', 'marketSubscription')
        return HistoricalStream(directory, listener)

    @staticmethod
    def create_historical_generator_stream(directory, listener=None):
        """
        Uses generator listener/cache to parse betfair
        historical data:
            https://historicdata.betfair.com/#/home

        :param str directory: Directory of betfair data
        :param BaseListener listener: Listener object

        :rtype: HistoricalGeneratorStream
        """
        listener = listener if listener else StreamListener()
        listener.register_stream('HISTORICAL', 'marketSubscription')
        return HistoricalGeneratorStream(directory, listener)



import requests
import datetime
from requests import ConnectionError

from ..exceptions import (
    APIError,
    InvalidResponse,
)
from ..utils import check_status_code
from ..compat import json

# monkeypatching requests
# https://github.com/kennethreitz/requests/issues/1595
requests.models.json = json


class BaseEndpoint(object):

    connect_timeout = 3.05
    read_timeout = 16
    _error = APIError

    def __init__(self, parent):
        """
        :param parent: API client.
        """
        self.client = parent

    def request(self, method, params, session):
        """
        :param str method: Betfair api-ng method to be used.
        :param dict params: Params to be used in request
        :param Session session: Requests session to be used, reduces latency.
        """
        session = session or self.client.session
        request = self.create_req(method, params)
        date_time_sent = datetime.datetime.utcnow()
        try:
            response = session.post(
                self.url,
                data=request,
                headers=self.client.request_headers,
                timeout=(self.connect_timeout, self.read_timeout)
            )
        except ConnectionError:
            raise APIError(None, method, params, 'ConnectionError')
        except Exception as e:
            raise APIError(None, method, params, e)
        elapsed_time = (datetime.datetime.utcnow()-date_time_sent).total_seconds()

        check_status_code(response)
        try:
            response_data = response.json()
        except ValueError:
            raise InvalidResponse(response.text)

        if self._error_handler:
            self._error_handler(response_data, method, params)
        return response_data, elapsed_time

    @staticmethod
    def create_req(method, params):
        """
        :param method: Betfair api-ng method to be used.
        :param params: Params to be used in request.
        :return: Json payload.
        """
        return json.dumps(
            {
                'jsonrpc': '2.0',
                'method': method,
                'params': params,
                'id': 1
            }
        )

    def _error_handler(self, response, method=None, params=None):
        """
        :param response: Json response.
        :param params: Params to be used in request.
        :param method: Betfair api-ng method to be used.
        :return: None if no error or _error raised.
        """
        if response.get('result'):
            return
        elif response.get('error'):
            raise self._error(response, method, params)

    def process_response(self, response_json, resource, elapsed_time, lightweight):
        """
        :param dict/list response_json: Response in dict format
        :param BaseResource resource: Resource data structure
        :param float elapsed_time: Elapsed time of request
        :param bool lightweight: If True will return dict not a resource (22x faster)
        """
        if isinstance(response_json, list):
            result = response_json
        else:
            result = response_json.get('result', response_json)

        if lightweight:
            return result
        elif self.client.lightweight and lightweight is not False:
            return result
        elif isinstance(result, list):
            try:
                return [resource(elapsed_time=elapsed_time, **x) for x in result]
            except TypeError:
                raise InvalidResponse(response=result)
        else:
            try:
                return resource(elapsed_time=elapsed_time, **result)
            except TypeError:
                raise InvalidResponse(response=result)

    @property
    def url(self):
        return '%s%s' % (self.client.api_uri, 'betting/json-rpc/v1')



import datetime

from ..resources import (
    BaseResource,
    MarketBook,
    CurrentOrders,
    MarketDefinition,
)
from ..enums import (
    StreamingOrderType,
    StreamingPersistenceType,
    StreamingSide,
    StreamingStatus,
)
from ..exceptions import CacheError


class Available(object):
    """
    Data structure to hold prices/traded amount,
    designed to be as quick as possible.
    """

    def __init__(self, prices, deletion_select, reverse=False):
        """
        :param list prices: Current prices
        :param int deletion_select: Used to decide if update should delete cache
        :param bool reverse: Used for sorting
        """
        self.prices = prices or []
        self.deletion_select = deletion_select
        self.reverse = reverse

        self.serialise = []
        self.sort()

    def sort(self):
        self.prices.sort(reverse=self.reverse)
        self.serialise = [
            {'price': volume[self.deletion_select-1], 'size': volume[self.deletion_select]} for volume in self.prices
        ]

    def clear(self):
        self.prices = []
        self.sort()

    def update(self, book_update):
        for book in book_update:
            for (count, trade) in enumerate(self.prices):
                if trade[0] == book[0]:
                    if book[self.deletion_select] == 0:
                        del self.prices[count]
                        break
                    else:
                        self.prices[count] = book
                        break
            else:
                if book[self.deletion_select] != 0:
                    # handles betfair bug, http://forum.bdp.betfair.com/showthread.php?t=3351
                    self.prices.append(book)
        self.sort()


class RunnerBook(object):

    def __init__(self, id, ltp=None, tv=None, trd=None, atb=None, batb=None, bdatb=None, atl=None, batl=None,
                 bdatl=None, spn=None, spf=None, spb=None, spl=None, hc=0):
        self.selection_id = id
        self.last_price_traded = ltp
        self.total_matched = tv
        self.traded = Available(trd, 1)
        self.available_to_back = Available(atb, 1, True)
        self.best_available_to_back = Available(batb, 2)
        self.best_display_available_to_back = Available(bdatb, 2)
        self.available_to_lay = Available(atl, 1)
        self.best_available_to_lay = Available(batl, 2)
        self.best_display_available_to_lay = Available(bdatl, 2)
        self.starting_price_back = Available(spb, 1)
        self.starting_price_lay = Available(spl, 1)
        self.starting_price_near = spn
        self.starting_price_far = spf
        self.handicap = hc

    def update_traded(self, traded_update):
        """:param traded_update: [price, size]
        """
        if not traded_update:
            self.traded.clear()
        else:
            self.traded.update(traded_update)

    def serialise_available_to_back(self):
        if self.available_to_back.prices:
            return self.available_to_back.serialise
        elif self.best_display_available_to_back.prices:
            return self.best_display_available_to_back.serialise
        elif self.best_available_to_back.prices:
            return self.best_available_to_back.serialise
        else:
            return []

    def serialise_available_to_lay(self):
        if self.available_to_lay.prices:
            return self.available_to_lay.serialise
        elif self.best_display_available_to_lay.prices:
            return self.best_display_available_to_lay.serialise
        elif self.best_available_to_lay.prices:
            return self.best_available_to_lay.serialise
        return []

    def serialise(self, runner_definition):
        return {
            'status': runner_definition.get('status'),
            'ex': {
                'tradedVolume': self.traded.serialise,
                'availableToBack': self.serialise_available_to_back(),
                'availableToLay': self.serialise_available_to_lay()
            },
            'sp': {
                'nearPrice': self.starting_price_near,
                'farPrice': self.starting_price_far,
                'backStakeTaken': self.starting_price_back.serialise,
                'layLiabilityTaken': self.starting_price_lay.serialise,
                'actualSP': runner_definition.get('bsp')
            },
            'adjustmentFactor': runner_definition.get('adjustmentFactor'),
            'removalDate': runner_definition.get('removalDate'),
            'lastPriceTraded': self.last_price_traded,
            'handicap': self.handicap,
            'totalMatched': self.total_matched,
            'selectionId': self.selection_id
        }


class MarketBookCache(BaseResource):

    def __init__(self, **kwargs):
        super(MarketBookCache, self).__init__(**kwargs)
        self.publish_time = kwargs.get('publish_time')
        self.market_id = kwargs.get('id')
        self.image = kwargs.get('img')
        self.total_matched = kwargs.get('tv')
        if 'marketDefinition' not in kwargs:
            raise CacheError('"EX_MARKET_DEF" must be requested to use cache')
        self.market_definition = kwargs['marketDefinition']

        self.runners = []
        self.runner_dict = {}
        self.market_definition_runner_dict = {}
        self._update_runner_dict()
        self._update_market_definition_runner_dict()

    def update_cache(self, market_change, publish_time):
        self._datetime_updated = self.strip_datetime(publish_time) or self._datetime_updated
        self.publish_time = publish_time

        if 'marketDefinition' in market_change:
            self.market_definition = market_change['marketDefinition']
            self._update_market_definition_runner_dict()

        if 'tv' in market_change:
            self.total_matched = market_change['tv']

        if 'rc' in market_change:
            for new_data in market_change['rc']:
                runner = self.runner_dict.get(
                    (new_data['id'], new_data.get('hc', 0))
                )
                if runner:
                    if 'ltp' in new_data:
                        runner.last_price_traded = new_data['ltp']
                    if 'tv' in new_data:  # if runner removed tv: 0 is returned
                        runner.total_matched = new_data['tv']
                    if 'spn' in new_data:
                        runner.starting_price_near = new_data['spn']
                    if 'spf' in new_data:
                        runner.starting_price_far = new_data['spf']
                    if 'trd' in new_data:
                        runner.update_traded(new_data['trd'])
                    if 'atb' in new_data:
                        runner.available_to_back.update(new_data['atb'])
                    if 'atl' in new_data:
                        runner.available_to_lay.update(new_data['atl'])
                    if 'batb' in new_data:
                        runner.best_available_to_back.update(new_data['batb'])
                    if 'batl' in new_data:
                        runner.best_available_to_lay.update(new_data['batl'])
                    if 'bdatb' in new_data:
                        runner.best_display_available_to_back.update(new_data['bdatb'])
                    if 'bdatl' in new_data:
                        runner.best_display_available_to_lay.update(new_data['bdatl'])
                    if 'spb' in new_data:
                        runner.starting_price_back.update(new_data['spb'])
                    if 'spl' in new_data:
                        runner.starting_price_lay.update(new_data['spl'])
                else:
                    self.runners.append(RunnerBook(**new_data))
                    self._update_runner_dict()

    def create_resource(self, unique_id, streaming_update, lightweight):
        data = self.serialise
        data["streaming_unique_id"] = unique_id
        data["streaming_update"] = streaming_update
        if lightweight:
            return data
        else:
            return MarketBook(
                elapsed_time=(datetime.datetime.utcnow()-self._datetime_updated).total_seconds(),
                market_definition=MarketDefinition(**self.market_definition),
                **data
            )

    def _update_runner_dict(self):
        self.runner_dict = {
            (runner.selection_id, runner.handicap): runner for runner in self.runners
        }

    def _update_market_definition_runner_dict(self):
        self.market_definition_runner_dict = {
            (runner['id'], runner.get('hc', 0)): runner for runner in self.market_definition['runners']
        }

    @property
    def serialise(self):
        """Creates standard market book json response,
        will error if EX_MARKET_DEF not incl.
        """
        return {
            'marketId': self.market_id,
            'totalAvailable': None,
            'isMarketDataDelayed': None,
            'lastMatchTime': None,
            'betDelay': self.market_definition.get('betDelay'),
            'version': self.market_definition.get('version'),
            'complete': self.market_definition.get('complete'),
            'runnersVoidable': self.market_definition.get('runnersVoidable'),
            'totalMatched': self.total_matched,
            'status': self.market_definition.get('status'),
            'bspReconciled': self.market_definition.get('bspReconciled'),
            'crossMatching': self.market_definition.get('crossMatching'),
            'inplay': self.market_definition.get('inPlay'),
            'numberOfWinners': self.market_definition.get('numberOfWinners'),
            'numberOfRunners': len(self.market_definition.get('runners')),
            'numberOfActiveRunners': self.market_definition.get('numberOfActiveRunners'),
            'runners': [
                runner.serialise(
                    self.market_definition_runner_dict[(runner.selection_id, runner.handicap)]
                ) for runner in self.runners
            ],
            'publishTime': self.publish_time,
            'priceLadderDefinition': self.market_definition.get('priceLadderDefinition'),
            'keyLineDescription': self.market_definition.get('keyLineDefinition'),
            'marketDefinition': self.market_definition,  # used in lightweight
        }


class UnmatchedOrder(object):

    def __init__(self, id, p, s, side, status, ot, pd, sm, sr, sl, sc, sv, rfo, rfs, pt=None,
                 md=None, avp=None, bsp=None, ld=None, rac=None, rc=None, lsrc=None, **kwargs):
        self.bet_id = id
        self.price = p
        self.size = s
        self.bsp_liability = bsp
        self.side = side
        self.status = status
        self.persistence_type = pt
        self.order_type = ot
        self.placed_date = BaseResource.strip_datetime(pd)
        self.placed_date_string = self.create_placed_date_string()
        self.matched_date = BaseResource.strip_datetime(md)
        self.matched_date_string = self.create_matched_date_string()
        self.average_price_matched = avp
        self.size_matched = sm
        self.size_remaining = sr
        self.size_lapsed = sl
        self.size_cancelled = sc
        self.size_voided = sv
        self.regulator_auth_code = rac
        self.regulator_code = rc
        self.reference_order = rfo
        self.reference_strategy = rfs
        self.lapsed_date = ld
        self.lapse_status_reason_code = lsrc  # todo add to output?

    def create_placed_date_string(self):
        if self.placed_date:
            return self.placed_date.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    def create_matched_date_string(self):
        if self.matched_date:
            return self.matched_date.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    def serialise(self, market_id, selection_id, handicap):
        return {
            'averagePriceMatched': self.average_price_matched or 0.0,
            'betId': self.bet_id,
            'bspLiability': self.bsp_liability,
            'handicap': handicap,
            'marketId': market_id,
            'matchedDate': self.matched_date_string,
            'orderType': StreamingOrderType[self.order_type].value,
            'persistenceType': StreamingPersistenceType[self.persistence_type].value if self.persistence_type else None,
            'placedDate': self.placed_date_string,
            'priceSize': {
                'price': self.price,
                'size': self.size
            },
            'regulatorCode': self.regulator_code,
            'selectionId': selection_id,
            'side': StreamingSide[self.side].value,
            'sizeCancelled': self.size_cancelled,
            'sizeLapsed': self.size_lapsed,
            'sizeMatched': self.size_matched,
            'sizeRemaining': self.size_remaining,
            'sizeVoided': self.size_voided,
            'status': StreamingStatus[self.status].value,
            'customerStrategyRef': self.reference_strategy,
            'customerOrderRef': self.reference_order,
        }


class OrderBookRunner(object):

    def __init__(self, id, fullImage=None, ml=None, mb=None, uo=None, hc=0, smc=None):
        self.selection_id = id
        self.full_image = fullImage
        self.matched_lays = Available(ml, 1)
        self.matched_backs = Available(mb, 1)
        self.unmatched_orders = {i["id"]: UnmatchedOrder(**i) for i in uo} if uo else {}
        self.handicap = hc
        self.strategy_matches = smc

    def update_unmatched(self, unmatched_orders):
        for unmatched_order in unmatched_orders:
            self.unmatched_orders[unmatched_order["id"]] = UnmatchedOrder(**unmatched_order)

    def serialise_orders(self, market_id):
        orders = list(self.unmatched_orders.values())  # order may be added (#232)
        return [
            order.serialise(market_id, self.selection_id, self.handicap) for order in orders
        ]


class OrderBookCache(BaseResource):

    def __init__(self, **kwargs):
        super(OrderBookCache, self).__init__(**kwargs)
        self.publish_time = kwargs.get('publish_time')
        self.market_id = kwargs.get('id')
        self.closed = kwargs.get('closed')
        self.runners = []

    def update_cache(self, order_book, publish_time):
        self._datetime_updated = self.strip_datetime(publish_time)
        self.publish_time = publish_time
        if "closed" in order_book:
            self.closed = order_book["closed"]

        for order_changes in order_book.get('orc', []):
            selection_id = order_changes['id']
            handicap = order_changes.get('hc', 0)
            runner = self.runner_dict.get(
                (selection_id, handicap)
            )
            if runner:
                if 'ml' in order_changes:
                    runner.matched_lays.update(order_changes['ml'])
                if 'mb' in order_changes:
                    runner.matched_backs.update(order_changes['mb'])
                if 'uo' in order_changes:
                    runner.update_unmatched(order_changes['uo'])
            else:
                self.runners.append(OrderBookRunner(**order_changes))

    def create_resource(self, unique_id, streaming_update, lightweight):
        data = self.serialise
        data["streaming_unique_id"] = unique_id
        data["streaming_update"] = streaming_update
        if lightweight:
            return data
        else:
            return CurrentOrders(
                elapsed_time=(datetime.datetime.utcnow()-self._datetime_updated).total_seconds(),
                publish_time=self.publish_time,
                **data
            )

    @property
    def runner_dict(self):
        return {(runner.selection_id, runner.handicap): runner for runner in self.runners}

    @property
    def serialise(self):
        orders = []
        for runner in self.runners:
            orders.extend(runner.serialise_orders(self.market_id))
        return {
            'currentOrders': orders,
            'moreAvailable': False
        }



import unittest

from requests.exceptions import ConnectionError

from betfairlightweight import APIClient
from betfairlightweight.endpoints.logininteractive import LoginInteractive
from betfairlightweight.exceptions import LoginError, APIError, InvalidResponse
from betfairlightweight.resources import LoginResource
from tests import mock
from tests.unit.tools import create_mock_json


class LoginInteractiveTest(unittest.TestCase):

    def setUp(self):
        client = APIClient('username', 'password', 'app_key', 'UK')
        self.login = LoginInteractive(client)

    @mock.patch('betfairlightweight.endpoints.logininteractive.LoginInteractive.request')
    def test_call(self, mock_response):
        mock_json = create_mock_json('tests/resources/login_interactive_success.json')
        mock_response.return_value = (mock_json.json(), 1.3)
        response = self.login()

        assert isinstance(response, LoginResource)
        assert self.login.client.session_token == mock_json.json().get('token')

    @mock.patch('betfairlightweight.baseclient.BaseClient.login_headers')
    @mock.patch('betfairlightweight.baseclient.requests.post')
    def test_request(self, mock_post, mock_login_headers):
        mock_response = create_mock_json('tests/resources/login_interactive_success.json')
        mock_post.return_value = mock_response

        url = 'https://identitysso.betfair.com/api/login'
        response = self.login.request()

        mock_post.assert_called_once_with(url, data={'username': 'username', 'password': 'password'},
                                          headers=mock_login_headers)
        assert response[0] == mock_response.json()

    @mock.patch('betfairlightweight.baseclient.BaseClient.login_headers')
    @mock.patch('betfairlightweight.baseclient.requests.post')
    def test_request_error(self, mock_post, mock_login_headers):
        mock_post.side_effect = ValueError()

        with self.assertRaises(APIError):
            self.login.request()

        mock_post.side_effect = ConnectionError()
        with self.assertRaises(APIError):
            self.login.request()

    @mock.patch('betfairlightweight.baseclient.BaseClient.login_headers')
    @mock.patch('betfairlightweight.baseclient.requests.post')
    def test_request_json_error(self, mock_post, mock_login_headers):
        mock_response = mock.Mock()
        mock_response.status_code = 200
        mock_response.json.side_effect = ValueError()
        mock_post.return_value = mock_response

        with self.assertRaises(InvalidResponse):
            self.login.request()

    def test_login_error_handler(self):
        mock_response = create_mock_json('tests/resources/login_interactive_success.json')
        assert self.login._error_handler(mock_response.json()) is None

        mock_response = create_mock_json('tests/resources/login_interactive_fail.json')
        with self.assertRaises(LoginError):
            self.login._error_handler(mock_response.json())

    def test_url(self):
        assert self.login.url == 'https://identitysso.betfair.com/api/login'

    def test_data(self):
        assert self.login.data == {'username': 'username', 'password': 'password'}



import datetime
import logging
import time

from .cache import (
    MarketBookCache,
    OrderBookCache,
)

logger = logging.getLogger(__name__)


class BaseStream(object):
    """Separate stream class to hold market/order caches
    """

    _lookup = 'mc'

    def __init__(self, listener):
        self._listener = listener

        self._initial_clk = None
        self._clk = None
        self._caches = {}
        self._updates_processed = 0
        self._on_creation()

        self.time_created = datetime.datetime.utcnow()
        self.time_updated = datetime.datetime.utcnow()

    def on_subscribe(self, data):
        self._update_clk(data)
        publish_time = data.get('pt')

        if self._lookup in data:
            self._process(data[self._lookup], publish_time)
        logger.info('[Stream: %s]: %s %s added' % (self.unique_id, len(self._caches), self._lookup))

    def on_heartbeat(self, data):
        self._update_clk(data)

    def on_resubscribe(self, data):
        self.on_update(data)
        logger.info('[Stream: %s]: %s %s resubscribed' % (self.unique_id, len(self._caches), self._lookup))

    def on_update(self, data):
        self._update_clk(data)

        publish_time = data['pt']
        latency = self._calc_latency(publish_time)
        if latency > self._max_latency:
            logger.warning('[Stream: %s]: Latency high: %s' % (self.unique_id, latency))

        if self._lookup in data:
            self._process(data[self._lookup], publish_time)

    def clear_cache(self):
        self._caches.clear()

    def snap(self, market_ids=None):
        return [
            cache.create_resource(self.unique_id, None, self._lightweight) for cache in list(self._caches.values())
            if market_ids is None or cache.market_id in market_ids
        ]

    def on_process(self, output):
        if self.output_queue:
            self.output_queue.put(output)

    def _on_creation(self):
        logger.info('[Stream: %s]: "%s" created' % (self.unique_id, self))

    def _process(self, book_data, publish_time):
        pass

    def _update_clk(self, data):
        (initial_clk, clk) = (data.get('initialClk'), data.get('clk'))
        if initial_clk:
            self._initial_clk = initial_clk
        if clk:
            self._clk = clk
        self.time_updated = datetime.datetime.utcnow()

    @property
    def unique_id(self):
        return self._listener.stream_unique_id

    @property
    def output_queue(self):
        return self._listener.output_queue

    @property
    def _max_latency(self):
        return self._listener.max_latency

    @property
    def _lightweight(self):
        return self._listener.lightweight

    @staticmethod
    def _calc_latency(publish_time):
        return time.time() - publish_time / 1e3

    def __len__(self):
        return len(self._caches)

    def __str__(self):
        return 'BaseStream'

    def __repr__(self):
        return '<BaseStream>'


class MarketStream(BaseStream):

    _lookup = 'mc'

    def _process(self, market_books, publish_time):
        output_market_book = []
        for market_book in market_books:
            market_id = market_book['id']
            market_book_cache = self._caches.get(market_id)

            if market_book.get('img') or market_book_cache is None:  # historic data does not contain img
                market_book_cache = MarketBookCache(publish_time=publish_time, **market_book)
                self._caches[market_id] = market_book_cache
                logger.info('[MarketStream: %s] %s added, %s markets in cache' %
                            (self.unique_id, market_id, len(self._caches)))

            market_book_cache.update_cache(market_book, publish_time)
            self._updates_processed += 1

            output_market_book.append(
                market_book_cache.create_resource(self.unique_id, market_book, self._lightweight)
            )
        self.on_process(output_market_book)

    def __str__(self):
        return 'MarketStream'

    def __repr__(self):
        return '<MarketStream [%s]>' % len(self)


class OrderStream(BaseStream):

    _lookup = 'oc'

    def _process(self, order_books, publish_time):
        output_order_book = []
        for order_book in order_books:
            market_id = order_book['id']
            order_book_cache = self._caches.get(market_id)

            if order_book_cache is None:
                order_book_cache = OrderBookCache(publish_time=publish_time, **order_book)
                self._caches[market_id] = order_book_cache
                logger.info('[OrderStream: %s] %s added, %s markets in cache' %
                            (self.unique_id, market_id, len(self._caches)))

            order_book_cache.update_cache(order_book, publish_time)
            self._updates_processed += 1

            output_order_book.append(
                order_book_cache.create_resource(self.unique_id, order_book, self._lightweight)
            )
        self.on_process(output_order_book)

    def __str__(self):
        return 'OrderStream'

    def __repr__(self):
        return '<OrderStream [%s]>' % len(self)



import unittest
from tests import mock
from requests.exceptions import ConnectionError

from betfairlightweight import APIClient
from betfairlightweight.endpoints.racecard import RaceCard
from betfairlightweight.exceptions import APIError, RaceCardError, InvalidResponse


class RaceCardTest(unittest.TestCase):

    def setUp(self):
        self.client = APIClient('username', 'password', 'app_key', 'UK')
        self.race_card = RaceCard(self.client)

    def test_init(self):
        assert self.race_card.connect_timeout == 3.05
        assert self.race_card._error == APIError
        assert self.race_card.client == self.client
        assert self.race_card.app_key is None

    def test_login(self):
        mock_session = mock.Mock()
        mock_response = mock.Mock()
        mock_response.text = '"appKey": "1234",'
        mock_session.get.return_value = mock_response
        self.race_card.login(mock_session)

        assert self.race_card.app_key == '1234'
        mock_session.get.assert_called_with(self.race_card.login_url)
        assert mock_session.get.call_count == 1

    def test_login_error(self):
        mock_session = mock.Mock()
        mock_response = mock.Mock()
        mock_response.text = '"applesKey": "1234",'
        mock_session.get.return_value = mock_response

        with self.assertRaises(RaceCardError):
            self.race_card.login(mock_session)

        assert self.race_card.app_key is None
        mock_session.get.assert_called_with(self.race_card.login_url)
        assert mock_session.get.call_count == 1

    def test_login_connection_error(self):
        mock_session = mock.Mock()
        mock_session.get.side_effect = ConnectionError()

        with self.assertRaises(APIError):
            self.race_card.login(mock_session)

    def test_login_unknown_error(self):
        mock_session = mock.Mock()
        mock_session.get.side_effect = ValueError()

        with self.assertRaises(APIError):
            self.race_card.login(mock_session)

    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.process_response')
    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.request', return_value=(mock.Mock(), 1.3))
    def test_get_race_card(self, mock_request, mock_process_response):
        market_ids = ['1', '2']
        data_entries = 'test'
        with self.assertRaises(RaceCardError):
            self.race_card.get_race_card(market_ids)

        self.race_card.app_key = '1234'
        self.race_card.get_race_card(market_ids=market_ids, data_entries=data_entries)

        mock_request.assert_called_once_with(
            "raceCard", session=None, params={'marketId': '1,2', 'dataEntries': 'test'})
        assert mock_request.call_count == 1

    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.process_response')
    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.request', return_value=(mock.Mock(), 1.3))
    def test_get_race_result(self, mock_request, mock_process_response):
        market_ids = ['1', '2']
        data_entries = 'test'
        with self.assertRaises(RaceCardError):
            self.race_card.get_race_result(market_ids)

        self.race_card.app_key = '1234'
        self.race_card.get_race_result(market_ids=market_ids, data_entries=data_entries)

        mock_request.assert_called_once_with(
            "raceResults", session=None, params={'marketId': '1,2', 'sortBy': 'DATE_DESC', 'dataEntries': 'test'})
        assert mock_request.call_count == 1

    @mock.patch('betfairlightweight.endpoints.racecard.check_status_code')
    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.create_req')
    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.headers')
    @mock.patch('betfairlightweight.baseclient.requests.get')
    def test_request(self, mock_get, mock_login_headers, mock_create_req, mock_check_status_code):
        mock_login_headers.return_value = {}

        self.race_card.request(method="test")
        _url = "https://www.betfair.com/rest/v2/test"

        mock_get.assert_called_with(_url, headers=mock_login_headers, params=None)
        assert mock_get.call_count == 1

    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.create_req')
    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.headers')
    @mock.patch('betfairlightweight.baseclient.requests.get')
    def test_request_error(self, mock_get, mock_login_headers, mock_create_req):
        mock_get.side_effect = ConnectionError()
        with self.assertRaises(APIError):
            self.race_card.request()

        mock_get.side_effect = ValueError()
        with self.assertRaises(APIError):
            self.race_card.request()

    @mock.patch('betfairlightweight.endpoints.racecard.check_status_code')
    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.create_req')
    @mock.patch('betfairlightweight.endpoints.racecard.RaceCard.headers')
    @mock.patch('betfairlightweight.baseclient.requests.get')
    def test_request_error(self, mock_get, mock_login_headers, mock_create_req, mock_check_status_code):
        response = mock.Mock()
        mock_get.return_value = response
        response.json.side_effect = ValueError()

        with self.assertRaises(InvalidResponse):
            self.race_card.request()

    def test_create_race_card_req(self):
        assert self.race_card.create_race_card_req(['1', '2'], None) == {
            'dataEntries': "RACE, TIMEFORM_DATA, RUNNERS, RUNNER_DETAILS",
            'marketId': '1,2'
        }
        assert self.race_card.create_race_card_req(['1', '2'], ['RACE']) == {
            'dataEntries': ['RACE'],
            'marketId': '1,2'
        }

    def test_create_race_result_req(self):
        assert self.race_card.create_race_result_req(['1', '2'], None) == {
            'dataEntries': "RUNNERS, MARKETS, PRICES, RACE, COURSE",
            'marketId': '1,2',
            'sortBy': 'DATE_DESC',
        }
        assert self.race_card.create_race_result_req(['1', '2'], ['RACE']) == {
            'dataEntries': ['RACE'],
            'marketId': '1,2',
            'sortBy': 'DATE_DESC',
        }

    def test_headers(self):
        assert self.race_card.headers == {
            'Connection': 'keep-alive',
            'Content-Type': 'application/json',
            'X-Application': None
        }

    def test_urls(self):
        assert self.race_card.url == 'https://www.betfair.com/rest/v2/'
        assert self.race_card.login_url == 'https://www.betfair.com/exchange/plus/'



import unittest
from tests import mock

from betfairlightweight import APIClient
from betfairlightweight.endpoints.account import Account
from betfairlightweight.exceptions import APIError


class AccountInit(unittest.TestCase):

    def test_base_endpoint_init(self):
        client = APIClient('username', 'password', 'app_key')
        account = Account(client)
        assert account.connect_timeout == 6.05
        assert account.read_timeout == 16
        assert account._error == APIError
        assert account.client == client
        assert account.URI == 'AccountAPING/v1.0/'


class AccountTest(unittest.TestCase):

    def setUp(self):
        client = APIClient('username', 'password', 'app_key', 'UK')
        self.account = Account(client)

    def test_init(self):
        assert self.account.URI == 'AccountAPING/v1.0/'
        assert self.account.connect_timeout == 6.05

    @mock.patch('betfairlightweight.endpoints.account.Account.process_response')
    @mock.patch('betfairlightweight.endpoints.account.Account.request', return_value=(mock.Mock(), mock.Mock()))
    def test_get_account_funds(self, mock_request, mock_process_response):
        self.account.get_account_funds()

        mock_request.assert_called_once_with('AccountAPING/v1.0/getAccountFunds', {}, None)
        assert mock_process_response.call_count == 1

    @mock.patch('betfairlightweight.endpoints.account.Account.process_response')
    @mock.patch('betfairlightweight.endpoints.account.Account.request', return_value=(mock.Mock(), mock.Mock()))
    def test_get_account_details(self, mock_request, mock_process_response):
        self.account.get_account_details()

        mock_request.assert_called_once_with('AccountAPING/v1.0/getAccountDetails', {}, None)
        assert mock_process_response.call_count == 1

    @mock.patch('betfairlightweight.endpoints.account.Account.process_response')
    @mock.patch('betfairlightweight.endpoints.account.Account.request', return_value=(mock.Mock(), mock.Mock()))
    def test_get_account_statement(self, mock_request, mock_process_response):
        self.account.get_account_statement()

        mock_request.assert_called_once_with('AccountAPING/v1.0/getAccountStatement', {'itemDateRange': {'from': None, 'to': None}}, None)
        assert mock_process_response.call_count == 1

    @mock.patch('betfairlightweight.endpoints.account.Account.process_response')
    @mock.patch('betfairlightweight.endpoints.account.Account.request', return_value=(mock.Mock(), mock.Mock()))
    def test_list_currency_rates(self, mock_request, mock_process_response):
        self.account.list_currency_rates()

        mock_request.assert_called_once_with('AccountAPING/v1.0/listCurrencyRates', {}, None)
        assert mock_process_response.call_count == 1

    def test_transfer_funds(self):
        with self.assertRaises(DeprecationWarning):
            self.account.transfer_funds()

    def test_url(self):
        assert self.account.url == '%s%s' % (self.account.client.api_uri, 'account/json-rpc/v1')



import datetime
from requests import ConnectionError

from ..exceptions import (
    APIError,
    InvalidResponse,
)
from ..utils import check_status_code
from .baseendpoint import BaseEndpoint
from .. import resources


class InPlayService(BaseEndpoint):
    """
    In play service operations.
    """

    def get_event_timeline(self, event_id, session=None, lightweight=None):
        """
        Returns event timeline for event id provided.

        :param int event_id: Event id to return
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: resources.EventTimeline
        """
        url = '%s%s' % (self.url, 'eventTimeline')
        params = {
            'eventId': event_id,
            'alt': 'json',
            'regionCode': 'UK',
            'locale': 'en_GB'
        }
        (response, elapsed_time) = self.request(params=params, session=session, url=url)
        return self.process_response(response, resources.EventTimeline, elapsed_time, lightweight)

    def get_event_timelines(self, event_ids, session=None, lightweight=None):
        """
        Returns a list of event timelines based on event id's
        supplied.

        :param list event_ids: List of event id's to return
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: list[resources.EventTimeline]
        """
        url = '%s%s' % (self.url, 'eventTimelines')
        params = {
            'eventIds': ','.join(str(x) for x in event_ids),
            'alt': 'json',
            'regionCode': 'UK',
            'locale': 'en_GB'
        }
        (response, elapsed_time) = self.request(params=params, session=session, url=url)
        return self.process_response(response, resources.EventTimeline, elapsed_time, lightweight)

    def get_scores(self, event_ids, session=None, lightweight=None):
        """
        Returns a list of scores based on event id's
        supplied.

        :param list event_ids: List of event id's to return
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: list[resources.Scores]
        """
        url = '%s%s' % (self.url, 'scores')
        params = {
            'eventIds': ','.join(str(x) for x in event_ids),
            'alt': 'json',
            'regionCode': 'UK',
            'locale': 'en_GB'
        }
        (response, elapsed_time) = self.request(params=params, session=session, url=url)
        return self.process_response(response, resources.Scores, elapsed_time, lightweight)

    def request(self, method=None, params=None, session=None, url=None):
        session = session or self.client.session
        date_time_sent = datetime.datetime.utcnow()
        try:
            response = session.get(url, params=params, headers=self.headers)
        except ConnectionError:
            raise APIError(None, method, params, 'ConnectionError')
        except Exception as e:
            raise APIError(None, method, params, e)
        elapsed_time = (datetime.datetime.utcnow() - date_time_sent).total_seconds()

        check_status_code(response)
        try:
            response_data = response.json()
        except ValueError:
            raise InvalidResponse(response.text)

        return response_data, elapsed_time

    @property
    def headers(self):
        return {
            'Connection': 'keep-alive',
            'Content-Type': 'application/json'
        }

    @property
    def url(self):
        return 'https://ips.betfair.com/inplayservice/v1.1/'



from .betfairstream import (
    BetfairStream,
    HistoricalStream,
    HistoricalGeneratorStream,
)
from .listener import (
    BaseListener,
    StreamListener,
)
from .stream import (
    MarketStream,
    OrderStream,
)



import unittest
from mock import Mock

from betfairlightweight.utils import (
    check_status_code,
    clean_locals,
    to_camel_case,
)
from betfairlightweight.exceptions import StatusCodeError


class UtilsTest(unittest.TestCase):

    def test_check_status_code_ok(self):
        resp = Mock()
        resp.status_code = 200

        assert check_status_code(resp) is None

    def test_check_status_code_fail(self):
        resp = Mock()
        resp.status_code = 400

        with self.assertRaises(StatusCodeError):
            check_status_code(resp)

    def test_api_request(self):
        pass

    def test_process_request(self):
        pass

    def test_convert_to_camel_case(self):
        assert to_camel_case('hello_world') == 'helloWorld'
        assert to_camel_case('async_') == 'async'


class UtilsTestCleanLocals(unittest.TestCase):

    def test_clean_locals(self, params=None, filter=123):
        params = clean_locals(locals())
        assert params == {'filter': 123}

    def test_clean_locals_params(self, params={'test': 456}, filter=123):
        params = clean_locals(locals())
        assert params == {'test': 456}
